//
// hello.step.45.c
//
// step response hello-world
//    9600 baud FTDI interface
//
// Neil Gershenfeld
// 10/27/10
//
// (c) Massachusetts Institute of Technology 2010
// Permission granted for experimental and personal use;
// license for commercial sale available from MIT.
//

#include <avr/io.h>
#include <util/delay.h>
//#include <stdint.h>
//#include <stdlib.h>
#include <avr/pgmspace.h>
#include <math.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 100 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define settle_delay() _delay_us(100) // settle delay
#define char_delay() _delay_ms(10) // char delay

#define tx2_port PORTD
#define rx2_port PORTB
#define tx2_dir DDRD
#define rx2_dir DDRB
#define tx2_pin (1<<PD7)
#define rx2_pin (1<<PB0)

#define tx1_port PORTD
#define rx1_port PORTD
#define tx1_dir DDRD
#define rx1_dir DDRD
#define tx1_pin (1<<PD5)
#define rx1_pin (1<<PD6)

#define tx_dir DDRB
#define tx_port PORTB
#define tx_pin (1 << PB3)

#define sense_port PORTC
#define sense_dir DDRC
#define charge_port PORTD
#define charge_dir DDRD

#define button_port PORTC
#define button_pin PINC
#define button_bv (1 << PC0)

#define NUM_POINTS 6
#define INIT_TIME 1
#define doDelay() _delay_us(15);
#define doInitDelay() _delay_us(1);
#define NUM_MODELS 2
#define NUM_PADS 4
#define REPEATS 16 //should be 2^SHIFT
#define SHIFT 4

#define DEBUG 0

static uint8_t sense_pins[4];//never changes, even though we use fewer
static uint8_t charge_pins[4];

static uint16_t model[NUM_POINTS*NUM_MODELS*NUM_PADS];//3
static uint16_t points[NUM_POINTS*NUM_PADS];
static uint32_t pointSums[NUM_POINTS*NUM_PADS];
static uint32_t dists[NUM_MODELS];
static float vars[NUM_POINTS*NUM_MODELS*NUM_PADS];
static uint8_t buttonPressedPrev;
static uint8_t numTrained;

static uint8_t preds[NUM_MODELS];
static uint32_t predThresh = 10000;//??
static uint8_t intruderCount = 0;

void init(){
//	output(tx1_dir, tx1_pin);//TX2 setup
//	set(tx1_port, tx1_pin);
//
//	output(tx2_dir, tx2_pin);//TX2 setup
//	set(tx2_port, tx2_pin);

//	output(tx1_port, tx1_pin);
//	set(tx1_port, tx1_pin);
	output(tx_dir, tx_pin);
	set(tx_port, tx_pin);

	output(tx1_dir, tx1_pin);
	output(tx2_dir, tx2_pin);
	output(rx1_dir, rx1_pin);
	output(rx2_dir, rx2_pin);

	set(button_port, button_bv);

	sense_pins[0] = (1<<PC5);
	sense_pins[1] = (1<<PC4);
	sense_pins[2] = (1<<PC3);
	sense_pins[3] = (1<<PC2);

	charge_pins[0] = (1<<PD0);
	charge_pins[1] = (1<<PD1);
	charge_pins[2] = (1<<PD2);
	charge_pins[3] = (1<<PD3);

	int i;
	for(i = 0; i < NUM_PADS; ++i){
		output(charge_dir, charge_pins[i]);
		clear(charge_port, charge_pins[i]);
	}
	settle_delay();
}

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
 }

void sendChar(char c){
	  put_char(&tx_port, tx_pin, c);
	  char_delay();
}

void sendCharOutput(char c){
	put_char(&tx2_port, tx2_pin, c);
	//put_char(&rx2_port, rx2_pin, c);
	char_delay();
}

//void sendCharDebug(char c){
//	put_char(&tx1_port, tx1_pin, c);
//	char_delay();
//}

/**
 * n should be the span of j (not i!!!!!!!!!!!)
 */
uint8_t getInd(uint8_t i, uint8_t j, uint8_t n){
	return i * n + j;
}

uint16_t getInd2(uint8_t i, uint8_t j, uint8_t k, uint8_t nj, uint8_t nk){
	uint16_t coef = nj;//force this to play nice
	coef *= nk;
	return i * coef + j * nk + k;
}

uint32_t getDist(uint32_t a, uint32_t b){
	int32_t a2, b2;
	a2 = a;
	b2 = b;
	a2 = (a2 - b2) * (a2 - b2);
	return (uint32_t)a2;
}

void delayNum(uint8_t n){
	//doInitDelay();
	uint8_t j = 0;
	for(j = 0; j < n; ++j)
		doDelay();
}

void setAD(uint8_t pad){
   //
   // init A/D
   //
	if(pad ==  0){
		   ADMUX = (0 << REFS1) | (1 << REFS0) // Vcc ref
			  | (0 << MUX3) | (1 << MUX2) | (0 << MUX1) | (1 << MUX0); // PC5
	} else if(pad == 1){
		 ADMUX = (0 << REFS1) | (1 << REFS0) // Vcc ref
			| (0 << MUX3) | (1 << MUX2) | (0 << MUX1) | (0 << MUX0); // PC4

	} else if(pad == 2){
		   ADMUX = (0 << REFS1) | (1 << REFS0) // Vcc ref
			  | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (1 << MUX0); // PC3
	} else if(pad == 3){
		   ADMUX = (0 << REFS1) | (1 << REFS0) // Vcc ref
			  | (0 << MUX3) | (0 << MUX2) | (1 << MUX1) | (0 << MUX0); // PC2
	}
}

void setAllPointsN(uint8_t pad){
	static unsigned char up_lo,up_hi,down_lo,down_hi;
	static uint32_t up, down;
	uint8_t i;
	setAD(pad);
	clear(charge_port, charge_pins[pad]);
	for(i = 0; i < NUM_POINTS; i++){
	   settle_delay();
	   set(charge_port, charge_pins[pad]);
	   delayNum(i);

	   PRR &= ~(1 << PRADC);
	   ADCSRA |= (1 << ADSC);
	   while (ADCSRA & (1 << ADSC));

	   up_lo = ADCL;
	   up_hi = ADCH;

	   settle_delay();
	   clear(charge_port, charge_pins[pad]);
	   //doDelay();
	   delayNum(i);

	   PRR &= ~(1 << PRADC);
		 ADCSRA |= (1 << ADSC);
		 while (ADCSRA & (1 << ADSC));

	     down_lo = ADCL;
	     down_hi = ADCH;

	     up = up_hi;
	     down = down_hi;

	     up = up << 8;
	     down = down << 8;

	     up += up_lo;
	     down += down_lo;
	     if(DEBUG){
		     uint32_t num = (up + (1023-down)) >> 1;
		     send32(num);
	     }
	     points[getInd(i,pad,NUM_PADS)] = (up + (1023-down)) >> 1;//average
	}
}

setAllPoints(){
	uint8_t i;
	for(i = 0; i < NUM_PADS; ++i){
		setAllPointsN(i);
	}
}

void setPointsViaSums(){
	uint8_t i,j,k;
	for(i = 0; i < NUM_POINTS*NUM_PADS; i++)
		pointSums[i] = 0;
	for(j = 0; j < REPEATS; ++j){//mean
		setAllPoints();
		for(i = 0; i < NUM_POINTS*NUM_PADS; i++)
			pointSums[i] += points[i];
	}
	for(i = 0; i < NUM_POINTS*NUM_PADS; i++)
		points[i] = pointSums[i] >> SHIFT;
}

void setVars(){//must have set model means.  will use numTrained
	uint8_t i,j,k;
	for(i = 0; i < NUM_POINTS; ++i)
		for(j = 0; j < NUM_PADS; ++j)
			vars[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] = 0.0f;
	for(i = 0; i < REPEATS; ++i){
		setAllPoints();
		for(j = 0; j < NUM_POINTS; ++j){
			for(k = 0; k < NUM_PADS; ++k)
				vars[getInd2(numTrained, j, k, NUM_POINTS, NUM_PADS)] += getDist(points[getInd(j,k,NUM_PADS)], model[getInd2(numTrained, j, k, NUM_POINTS,NUM_PADS)]);
		}
	}
	for(i = 0; i < NUM_POINTS; ++i)
		for(j = 0; j < NUM_PADS; ++j){
			float f = vars[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] / REPEATS;
			vars[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] = sqrt(f);
			if(vars[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] == 0.0f)
				vars[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] = 1.0f;
		}
}

void train(){
	uint8_t i,j;
	if(numTrained < NUM_MODELS){
		setPointsViaSums();
		for(i = 0; i < NUM_POINTS; i++)//simplest thing for now
			for(j = 0; j < NUM_PADS; ++j)
				model[getInd2(numTrained, i, j, NUM_POINTS, NUM_PADS)] = points[getInd(i, j, NUM_PADS)];
		setVars();
		++numTrained;
	}
}



void setDists(){
	uint32_t tempDist;
	setPointsViaSums();
	uint8_t i,j,k;
	for(i = 0; i < NUM_MODELS; ++i)
		dists[i] = 1;
	for(i = 0; i < numTrained; i++){
		tempDist = 0;
		for(j = 0; j < NUM_POINTS; ++j)
			for(k = 0; k < NUM_PADS; ++k)
				tempDist += getDist(points[getInd(j, k, NUM_PADS)], model[getInd2(i,j,k,NUM_POINTS, NUM_PADS)]);
		dists[i] = tempDist;
	}
}

void setDistsVar(){
	uint32_t tempDists[NUM_POINTS*NUM_PADS];
	uint32_t tempDist;
	uint8_t i,j,k,l;
	for(i = 0; i < NUM_MODELS; ++i)//init
		dists[i] = 10000000;

	for(i = 0; i < numTrained; ++i){
		for(j = 0; j < NUM_POINTS*NUM_PADS; ++j)
			tempDists[j] = 0;
		for(j = 0; j < REPEATS; ++j){
			setAllPoints();
			for(k = 0; k < NUM_POINTS; ++k)//sum SqDists, add variance as approximation to gaussian (laweoafsdf)
				for(l = 0; l < NUM_PADS; ++l)
					tempDists[getInd(k,l,NUM_PADS)] += getDist(points[getInd(k,l,NUM_PADS)], model[getInd2(i, k, l, NUM_POINTS, NUM_PADS)]);
		}
		for(j = 0; j < NUM_POINTS; ++j)//divide sqDist by var after the sum (avoid underflow)
			for(k = 0; k < NUM_PADS; ++k)
				tempDists[getInd(j, k, NUM_PADS)] /= vars[getInd2(i, j, k, NUM_POINTS, NUM_PADS)];
		tempDist = 0;
		for(j = 0; j < NUM_POINTS; ++j)
			for(k = 0; k < NUM_PADS; ++k)
				tempDist += tempDists[getInd(j,k,NUM_PADS)];
		dists[i] = tempDist;//could shift, but that'd just lose resolution

	}
}

unsigned char getClosest(){
	setDists();
	uint32_t dist;
	dist = UINT32_MAX;
	char closest = 12;
	uint8_t i;
	for(i = 0; i < numTrained; i++){
		if(dists[i] < dist){
			dist = dists[i];
			closest = i;
		}
	}
	return closest;
}

uint8_t buttonPressed(){
	if(button_pin & button_bv){
		return 0;
	}
	return 1;
}

void setPreds(){
	uint8_t i,j;
	//sendChar(4);
	//sendChar(4);
	//setDistsVar();
	setDists();
	for(i = 0; i < NUM_MODELS; ++i){
		preds[i] = 0;
	}
	for(i = 0; i < numTrained; ++i){
		send32(dists[i]);
		if(dists[i] < predThresh)
			preds[i] = 1;
	}
}

/**
 * returns 0 when there is an intruder;
 */
uint8_t checkIntruder(){
	if(numTrained < 2) return 1;
	setPreds();
	//sendModels();
	if(DEBUG){
		sendCharDebug(1);
		sendCharDebug(2);
		sendCharDebug(3);
		sendCharDebug(4);
		uint8_t i;
		for(i = 0; i < NUM_MODELS; ++i)
			sendCharDebug(preds[i]);
	}
	uint8_t recs = 0;
	uint8_t i;
	for(i = 0; i < numTrained; ++i)
		recs += preds[i];
	return recs;
}

void setOutputs(){
	set(tx1_port, tx1_pin);
	set(tx2_port, tx2_pin);
	set(rx1_port, rx1_pin);
	set(rx2_port, rx2_pin);
}

void clearOutputs(){
	clear(tx1_port, tx1_pin);
	clear(tx2_port, tx2_pin);
	clear(rx1_port, rx1_pin);
	clear(rx2_port, rx2_pin);
}

void fireAlarm(){
	//sendCharOutput(1);
	setOutputs();
	int c;
	for(c = 0; c < 10; ++c){
		_delay_ms(100);
	}
	clearOutputs();

}

//void send16(uint16_t n){
//	sendChar(n >> 8);
//	sendChar(n);
//}
//
void send32(uint32_t n){
	sendChar(n >> 24);
	sendChar(n >> 16);
	sendChar(n >> 8);
	sendChar(n);
}
//
//void sendData(){
//	uint8_t i;
//	for(i = 0; i < NUM_POINTS; i++){
//		send16(points[i]);
//	}
//}

void sendModels(){
	uint8_t i,j,k;
	for(i = 0; i < numTrained; i++){
		for(j = 0; j < NUM_POINTS; ++j){
			for(k = 0; k < NUM_PADS; ++k)
				send32(model[getInd2(i,j,k,NUM_POINTS,NUM_PADS)]);
		}
	}
}

//void sendDists(){
//	uint8_t i;
//	for(i = 0; i < NUM_MODELS; ++i)
//	{
//		send32(dists[i]);
//	}
//}
//
//void debug(){
//	  sendChar(getClosest());
//	  sendData();
////	  sendModels();
//}

int main(void) {
	init();
	//fireAlarm();
	buttonPressedPrev = 0;
	numTrained = 0;
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);


   PRR &= ~(1 << PRADC);
   //AD enable
   ADCSRA = (1 << ADEN) // enable
	  | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128

   uint8_t i;
   for(i = 0; i < NUM_MODELS; ++i)
	   dists[i] = 1;


   //fireAlarm();
   //
   // main loop
   //
   train();//0 is always background
   //train();

   while (1) {

//	   if(buttonPressed()){
//		   //clear(rx2_port, rx2_pin);
//		   set(tx2_port, tx2_pin);
//
//	   } else {
//		   //set(rx2_port, rx2_pin);
//		   clear(tx2_port, tx2_pin);
//	   }
	  if(buttonPressed()){
		  if(!buttonPressedPrev){
			  buttonPressedPrev = 1;
			  train();
			  fireAlarm();
		  }
	  } else {
		  buttonPressedPrev = 0;
		  if(checkIntruder()==0){
			  intruderCount++;
		  } else {
			  intruderCount = 0;
		  }
		  if(intruderCount > 5){
			  fireAlarm();
			  intruderCount = 0;
		  }
	  }

   }
}
